@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,31 +24,28 @@
 # Invariants: The periodic callback assumes pcm_ok. If we don't have
 # at least pcm_ok, we have no sound whatsoever.
 
+from __future__ import print_function
+
 import renpy.audio  # @UnusedImport
 import renpy.display  # @UnusedImport
+from renpy import six
 
 import time
 import pygame_sdl2  # @UnusedImport
 import os
+import re
+import threading
+import sys
 
 # Import the appropriate modules, or set them to None if we cannot.
 
 disable = os.environ.get("RENPY_DISABLE_SOUND", "")
 
-if 'pss' not in disable:
+if not disable:
+    import renpy.audio.renpysound as renpysound
+else:
+    renpysound = None
 
-    try:
-        import pysdlsound as pss
-        pss.check_version(4)  # @UndefinedVariable
-    except:
-        import traceback
-        traceback.print_exc()
-        pss = None
-
-# Save the mixer, and restore it at exit.
-
-old_wave = None
-old_midi = None
 
 # This is True if we were able to sucessfully enable the pcm audio.
 pcm_ok = None
@@ -56,6 +53,7 @@ pcm_ok = None
 unique = time.time()
 serial = 0
 
+
 def get_serial():
     """
     Gets a globally unique serial number for each music change.
@@ -80,10 +78,11 @@ class QueueEntry(object):
     A queue entry object.
     """
 
-    def __init__(self, filename, fadein, tight):
+    def __init__(self, filename, fadein, tight, loop):
         self.filename = filename
         self.fadein = fadein
         self.tight = tight
+        self.loop = loop
 
 
 class MusicContext(renpy.python.RevertableObject):
@@ -95,6 +94,8 @@ class MusicContext(renpy.python.RevertableObject):
 
     __version__ = 0
 
+    pause = False
+
     def __init__(self):
 
         super(MusicContext, self).__init__()
@@ -123,6 +124,9 @@ class MusicContext(renpy.python.RevertableObject):
         # Should we force stop this channel?
         self.force_stop = False
 
+        # Should we pause this channel?
+        self.pause = False
+
     def copy(self):
         """
         Returns a shallow copy of this context.
@@ -133,15 +137,20 @@ class MusicContext(renpy.python.RevertableObject):
 
         return rv
 
+
 # The next channel number to be assigned.
 next_channel_number = 0
 
+# the lock that mediates between the periodic and main threads.
+lock = threading.RLock()
+
+
 class Channel(object):
     """
     This stores information about the currently-playing music.
     """
 
-    def __init__(self, name, default_loop, stop_on_mute, tight, file_prefix, file_suffix, buffer_queue):
+    def __init__(self, name, default_loop, stop_on_mute, tight, file_prefix, file_suffix, buffer_queue, movie, framedrop):
 
         # The name assigned to this channel. This is used to look up
         # information about the channel in the MusicContext object.
@@ -212,6 +221,9 @@ class Channel(object):
         # Should we buffer upcoming music/video in the queue?
         self.buffer_queue = buffer_queue
 
+        # Are we paused?
+        self.paused = None
+
         if default_loop is None:
             # By default, should we loop the music?
             self.default_loop = True
@@ -222,6 +234,15 @@ class Channel(object):
             self.default_loop = default_loop
             self.default_loop_set = True
 
+        # Is this a movie channel?
+
+        if movie:
+            if framedrop:
+                self.movie = renpy.audio.renpysound.DROP_VIDEO
+            else:
+                self.movie = renpy.audio.renpysound.NODROP_VIDEO
+        else:
+            self.movie = renpy.audio.renpysound.NO_VIDEO
 
     def get_number(self):
         """
@@ -255,6 +276,58 @@ class Channel(object):
 
     context = property(get_context)
 
+    def split_filename(self, filename, looped):
+        """
+        Splits a filename into a filename, start time, and end time.
+        """
+
+        def exception(msg):
+            return Exception("Parsing audio spec {!r}: {}.".format(filename, msg))
+
+        def expect_float():
+            if not spec:
+                raise exception("expected float at end.")
+
+            v = spec.pop(0)
+
+            try:
+                return float(v)
+            except:
+                raise exception("expected float, got {!r}.".format(v))
+
+        m = re.match(r'<(.*)>(.*)', filename)
+        if not m:
+            return filename, 0, -1
+
+        spec = m.group(1)
+        fn = m.group(2)
+
+        spec = spec.split()
+
+        start = 0
+        loop = None
+        end = -1
+
+        while spec:
+            clause = spec.pop(0)
+
+            if clause == "from":
+                start = expect_float()
+            elif clause == "to":
+                end = expect_float()
+            elif clause == "loop":
+                loop = expect_float()
+            elif clause == "silence":
+                end = expect_float()
+                fn = "_silence.ogg"
+
+            else:
+                raise exception("expected keyword, got {!r}.".format(clause))
+
+        if (loop is not None) and looped:
+            start = loop
+
+        return fn, start, end
 
     def periodic(self):
         """
@@ -263,23 +336,23 @@ class Channel(object):
         """
 
         # Update the channel volume.
-        vol = self.chan_volume * renpy.game.preferences.volumes[self.mixer]
+        vol = self.chan_volume * renpy.game.preferences.volumes.get(self.mixer, 1.0)
 
         if vol != self.actual_volume:
-            pss.set_volume(self.number, vol)
+            renpysound.set_volume(self.number, vol)
             self.actual_volume = vol
 
-
         # This should be set from something that checks to see if our
         # mixer is muted.
-        force_stop = self.context.force_stop or (renpy.game.preferences.mute[self.mixer] and self.stop_on_mute)
+        force_stop = self.context.force_stop or (renpy.game.preferences.mute.get(self.mixer, False) and self.stop_on_mute)
 
         if self.playing and force_stop:
-            pss.stop(self.number)
+            renpysound.stop(self.number)
             self.playing = False
-            self.wait_stop = False
 
         if force_stop:
+            self.wait_stop = False
+
             if self.loop:
                 self.queue = self.queue[-len(self.loop):]
             else:
@@ -289,12 +362,14 @@ class Channel(object):
         # Should we do the callback?
         do_callback = False
 
+        topq = None
+
         # This has been modified so we only queue a single sound file
         # per call, to prevent memory leaks with really short sound
         # files. So this loop will only execute once, in practice.
         while True:
 
-            depth = pss.queue_depth(self.number)
+            depth = renpysound.queue_depth(self.number)
 
             if depth == 0:
                 self.wait_stop = False
@@ -320,7 +395,7 @@ class Channel(object):
                 break
 
             # If the queue is full, return.
-            if pss.queue_depth(self.number) >= 2:
+            if renpysound.queue_depth(self.number) >= 2:
                 break
 
             # Otherwise, we might be able to enqueue something.
@@ -337,12 +412,19 @@ class Channel(object):
                 continue
 
             try:
-                topf = load(self.file_prefix + topq.filename + self.file_suffix)
+                filename, start, end = self.split_filename(topq.filename, topq.loop)
+
+                if (end >= 0) and ((end - start) <= 0) and self.queue:
+                    continue
+
+                topf = load(self.file_prefix + filename + self.file_suffix)
+
+                renpysound.set_video(self.number, self.movie)
 
                 if depth == 0:
-                    pss.play(self.number, topf, topq.filename, paused=self.synchro_start, fadein=topq.fadein, tight=topq.tight)
+                    renpysound.play(self.number, topf, topq.filename, paused=self.synchro_start, fadein=topq.fadein, tight=topq.tight, start=start, end=end)
                 else:
-                    pss.queue(self.number, topf, topq.filename, fadein=topq.fadein, tight=topq.tight)
+                    renpysound.queue(self.number, topf, topq.filename, fadein=topq.fadein, tight=topq.tight, start=start, end=end)
 
                 self.playing = True
 
@@ -362,14 +444,29 @@ class Channel(object):
 
         if self.loop and not self.queue:
             for i in self.loop:
-                newq = QueueEntry(i, 0, topq.tight)
+                if topq is not None:
+                    newq = QueueEntry(i, 0, topq.tight, True)
+                else:
+                    newq = QueueEntry(i, 0, False, True)
+
                 self.queue.append(newq)
         else:
             do_callback = True
 
         # Queue empty callback.
         if do_callback and self.callback:
-            self.callback() # E1102
+            self.callback()  # E1102
+
+        want_pause = self.context.pause or global_pause
+
+        if self.paused != want_pause:
+
+            if want_pause:
+                self.pause()
+            else:
+                self.unpause()
+
+            self.paused = want_pause
 
     def dequeue(self, even_tight=False):
         """
@@ -379,14 +476,18 @@ class Channel(object):
         left in the queue unless all is given.
         """
 
-        self.queue = self.queue[:self.keep_queue]
-        self.loop = [ ]
+        with lock:
 
-        if not pcm_ok:
-            return
+            self.queue = self.queue[:self.keep_queue]
+            self.loop = [ ]
 
-        if self.keep_queue == 0:
-            pss.dequeue(self.number, even_tight)
+            if not pcm_ok:
+                return
+
+            if self.keep_queue == 0:
+                renpysound.dequeue(self.number, even_tight)
+                self.wait_stop = False
+                self.synchro_start = False
 
     def interact(self):
         """
@@ -399,20 +500,18 @@ class Channel(object):
 
             if self.pan_time != self.context.pan_time:
                 self.pan_time = self.context.pan_time
-                pss.set_pan(self.number,
-                            self.context.pan,
-                            0)
-
+                renpysound.set_pan(self.number,
+                                   self.context.pan,
+                                   0)
 
             if self.secondary_volume_time != self.context.secondary_volume_time:
                 self.secondary_volume_time = self.context.secondary_volume_time
-                pss.set_secondary_volume(self.number,
-                                         self.context.secondary_volume,
-                                         0)
+                renpysound.set_secondary_volume(self.number,
+                                                self.context.secondary_volume,
+                                                0)
 
         if not self.queue and self.callback:
-            self.callback() # E1102
-
+            self.callback()  # E1102
 
     def fadeout(self, secs):
         """
@@ -420,57 +519,68 @@ class Channel(object):
         of seconds. Also clears any queued music.
         """
 
-        self.keep_queue = 0
-        self.dequeue()
+        with lock:
 
-        if not pcm_ok:
-            return
+            self.keep_queue = 0
+            self.dequeue()
 
-        if secs == 0:
-            pss.stop(self.number)
-        else:
-            pss.fadeout(self.number, int(secs * 1000))
+            if not pcm_ok:
+                return
 
-    def enqueue(self, filenames, loop=True, synchro_start=False, fadein=0, tight=None):
+            if secs == 0:
+                renpysound.stop(self.number)
+            else:
+                renpysound.fadeout(self.number, secs)
 
-        for filename in filenames:
-            renpy.game.persistent._seen_audio[filename] = True  # @UndefinedVariable
+    def enqueue(self, filenames, loop=True, synchro_start=False, fadein=0, tight=None, loop_only=False):
 
-        if not pcm_ok:
-            return
+        with lock:
 
-        if tight is None:
-            tight = self.tight
+            for filename in filenames:
+                filename, _, _ = self.split_filename(filename, False)
+                renpy.game.persistent._seen_audio[filename] = True  # @UndefinedVariable
 
-        self.keep_queue += 1
+            if not pcm_ok:
+                return
 
-        for filename in filenames:
-            qe = QueueEntry(filename, int(fadein * 1000), tight)
-            self.queue.append(qe)
+            if not loop_only:
 
-            # Only fade the first thing in.
-            fadein = 0
+                if tight is None:
+                    tight = self.tight
 
-        if loop:
-            self.loop = list(filenames)
-        else:
-            self.loop = [ ]
+                self.keep_queue += 1
 
-        self.wait_stop = synchro_start
-        self.synchro_start = synchro_start
+                for filename in filenames:
+                    qe = QueueEntry(filename, int(fadein * 1000), tight, False)
+                    self.queue.append(qe)
+
+                    # Only fade the first thing in.
+                    fadein = 0
+
+                self.wait_stop = synchro_start
+                self.synchro_start = synchro_start
+
+            if loop:
+                self.loop = list(filenames)
+            else:
+                self.loop = [ ]
 
     def get_playing(self):
 
         if not pcm_ok:
             return None
 
-        rv = pss.playing_name(self.number)
+        rv = renpysound.playing_name(self.number)
 
-        if rv is None and self.queue:
-            rv = self.queue[0].filename
+        with lock:
 
-        if rv is None and self.loop:
-            rv = self.loop[0]
+            rv = renpysound.playing_name(self.number)
+
+            if rv is None and self.queue:
+                rv = self.queue[0].filename
+
+            if rv is None and self.loop:
+                rv = self.loop[0]
 
         return rv
 
@@ -482,36 +592,60 @@ class Channel(object):
         if not pcm_ok:
             return -1
 
-        return pss.get_pos(self.number)
+        return renpysound.get_pos(self.number)
+
+    def get_duration(self):
+
+        if not pcm_ok:
+            return 0.0
+
+        return renpysound.get_duration(self.number)
 
     def set_pan(self, pan, delay):
-        now = get_serial()
-        self.context.pan_time = now
-        self.context.pan = pan
 
-        if pcm_ok:
-            self.pan_time = self.context.pan_time
-            pss.set_pan(self.number, self.context.pan, delay)
+        with lock:
+
+            now = get_serial()
+            self.context.pan_time = now
+            self.context.pan = pan
+
+            if pcm_ok:
+                self.pan_time = self.context.pan_time
+                renpysound.set_pan(self.number, self.context.pan, delay)
 
     def set_secondary_volume(self, volume, delay):
-        now = get_serial()
-        self.context.secondary_volume_time = now
-        self.context.secondary_volume = volume
 
-        if pcm_ok:
-            self.secondary_volume_time = self.context.secondary_volume_time
-            pss.set_secondary_volume(self.number, self.context.secondary_volume, delay)
+        with lock:
+
+            now = get_serial()
+            self.context.secondary_volume_time = now
+            self.context.secondary_volume = volume
+
+            if pcm_ok:
+                self.secondary_volume_time = self.context.secondary_volume_time
+                renpysound.set_secondary_volume(self.number, self.context.secondary_volume, delay)
 
     def pause(self):
-        pss.pause(self.number)
+        with lock:
+            renpysound.pause(self.number)
 
     def unpause(self):
-        pss.unpause(self.number)
+        with lock:
+            renpysound.unpause(self.number)
 
+    def read_video(self):
+        if pcm_ok:
+            return renpysound.read_video(self.number)
+
+        return None
+
+    def video_ready(self):
+
+        if not pcm_ok:
+            return 1
+
+        return renpysound.video_ready(self.number)
 
-################################################################################
-# Android VideoPlayer Channel
-################################################################################
 
 # Use unconditional imports so these files get compiled during the build
 # process.
@@ -533,7 +667,7 @@ all_channels = [ ]
 channels = { }
 
 
-def register_channel(name, mixer=None, loop=None, stop_on_mute=True, tight=False, file_prefix="", file_suffix="", buffer_queue=True):
+def register_channel(name, mixer=None, loop=None, stop_on_mute=True, tight=False, file_prefix="", file_suffix="", buffer_queue=True, movie=False, framedrop=True):
     """
     :doc: audio
 
@@ -569,9 +703,20 @@ def register_channel(name, mixer=None, loop=None, stop_on_mute=True, tight=False
     `buffer_queue`
         Should we buffer the first second or so of a queued file? This should
         be True for audio, and False for movie playback.
+
+    `movie`
+        If true, this channel will be set up to play back videos.
+
+    `framedrop`
+        This controls what a video does when lagging. If true, frames will
+        be dropped to keep up with realtime and the soundtrack. If false,
+        Ren'Py will display frames late rather than dropping them.
     """
 
-    if not renpy.game.context().init_phase:
+    if name == "movie":
+        movie = True
+
+    if not renpy.game.context().init_phase and (" " not in name):
         raise Exception("Can't register channel outside of init phase.")
 
     if renpy.android and renpy.config.hw_video and name == "movie":
@@ -579,7 +724,7 @@ def register_channel(name, mixer=None, loop=None, stop_on_mute=True, tight=False
     elif renpy.ios and renpy.config.hw_video and name == "movie":
         c = IOSVideoChannel(name, default_loop=loop, file_prefix=file_prefix, file_suffix=file_suffix)
     else:
-        c = Channel(name, loop, stop_on_mute, tight, file_prefix, file_suffix, buffer_queue)
+        c = Channel(name, loop, stop_on_mute, tight, file_prefix, file_suffix, buffer_queue, movie=movie, framedrop=framedrop)
 
     c.mixer = mixer
 
@@ -597,16 +742,64 @@ def alias_channel(name, newname):
 
 def get_channel(name):
 
-    rv = channels.get(name)
+    rv = channels.get(name, None)
+
     if rv is None:
-        raise Exception("Audio channel %r is unknown." % name)
+
+        # Do we want to auto-define a new channel?
+        if name in renpy.config.auto_channels:
+
+            i = 0
+
+            while True:
+                c = get_channel("{} {}".format(name, i))
+
+                if not c.get_playing():
+                    return c
+
+                # Limit to one channel while skipping, to prevent sounds from
+                # piling up.
+                if renpy.config.skipping:
+                    return c
+
+                i += 1
+
+        # One of the channels that was just defined.
+        elif " " in name:
+
+            base = name.split()[0]
+            mixer, file_prefix, file_suffix = renpy.config.auto_channels[base]
+
+            register_channel(
+                name,
+                loop=False,
+                mixer=mixer,
+                file_prefix=file_prefix,
+                file_suffix=file_suffix,
+                )
+
+            return channels[name]
+
+        else:
+            raise Exception("Audio channel %r is unknown." % name)
 
     return rv
 
+
 def set_force_stop(name, value):
     get_channel(name).context.force_stop = value
 
+
+# The thread that call periodic.
+periodic_thread = None
+
+# True if we need it to quit.
+periodic_thread_quit = True
+
+
 def init():
+    global periodic_thread
+    global periodic_thread_quit
 
     global pcm_ok
     global mix_ok
@@ -616,14 +809,14 @@ def init():
         mix_ok = False
         return
 
-    if pcm_ok is None and pss:
+    if pcm_ok is None and renpysound:
         bufsize = 2048
 
         if 'RENPY_SOUND_BUFSIZE' in os.environ:
             bufsize = int(os.environ['RENPY_SOUND_BUFSIZE'])
 
         try:
-            pss.init(renpy.config.sound_sample_rate, 2, bufsize, False)
+            renpysound.init(renpy.config.sound_sample_rate, 2, bufsize, False)
             pcm_ok = True
         except:
             if renpy.config.debug_sound:
@@ -643,12 +836,31 @@ def init():
         renpy.game.preferences.volumes.setdefault(m, default_volume)
         renpy.game.preferences.mute.setdefault(m, False)
 
+    with periodic_condition:
 
-def quit(): #@ReservedAssignment
+        periodic_thread_quit = False
+
+        periodic_thread = threading.Thread(target=periodic_thread_main)
+        periodic_thread.daemon = True
+        periodic_thread.start()
+
+
+def quit():  # @ReservedAssignment
+
+    global periodic_thread
+    global periodic_thread_quit
 
     global pcm_ok
     global mix_ok
 
+    if periodic_thread is not None:
+        with periodic_condition:
+
+            periodic_thread_quit = True
+            periodic_condition.notify()
+
+        periodic_thread.join()
+
     if not pcm_ok:
         return
 
@@ -663,17 +875,20 @@ def quit(): #@ReservedAssignment
         c.wait_stop = False
         c.synchro_start = False
 
-    pss.quit()
+    renpysound.quit()
 
     pcm_ok = None
     mix_ok = None
 
+
+
 # The last-set pcm volume.
 pcm_volume = None
 
 old_emphasized = False
 
-def periodic():
+
+def periodic_pass():
     """
     The periodic sound callback. This is called at around 20hz, and is
     responsible for adjusting the volume of the playing music if
@@ -687,7 +902,6 @@ def periodic():
     if not pcm_ok:
         return False
 
-
     try:
 
         # A list of emphasized channels.
@@ -701,6 +915,9 @@ def periodic():
             if c.get_playing():
                 emphasized = True
 
+        if not renpy.game.preferences.emphasize_audio:
+            emphasized = False
+
         if emphasized and not old_emphasized:
             vol = renpy.config.emphasize_audio_volume
         elif not emphasized and old_emphasized:
@@ -708,9 +925,6 @@ def periodic():
         else:
             vol = None
 
-        if not renpy.game.preferences.emphasize_audio:
-            emphasized = False
-
         old_emphasized = emphasized
 
         if vol is not None:
@@ -723,7 +937,7 @@ def periodic():
         for c in all_channels:
             c.periodic()
 
-        pss.periodic()
+        renpysound.periodic()
 
         # Perform a synchro-start if necessary.
         need_ss = False
@@ -738,7 +952,7 @@ def periodic():
                 need_ss = True
 
         if need_ss:
-            pss.unpause_all()
+            renpysound.unpause_all()
 
             for c in all_channels:
                 c.synchro_start = False
@@ -748,6 +962,66 @@ def periodic():
             raise
 
 
+# The exception that's been thrown by the periodic thread.
+periodic_exc = None
+
+
+# Should we run the periodic thread now?
+run_periodic = False
+
+# The condition the perodic thread runs on.
+periodic_condition = threading.Condition()
+
+
+def periodic_thread_main():
+
+    global periodic_exc
+    global run_periodic
+
+    while True:
+        with periodic_condition:
+            if not run_periodic:
+                periodic_condition.wait(.05)
+
+            if periodic_thread_quit:
+                return
+
+            if not run_periodic:
+                continue
+
+            run_periodic = False
+
+        with lock:
+
+            try:
+                periodic_pass()
+            except Exception:
+                periodic_exc = sys.exc_info()
+
+
+def periodic():
+    global periodic_exc
+    global run_periodic
+
+    if not renpy.config.audio_periodic_thread:
+        periodic_pass()
+        return
+
+    with periodic_condition:
+
+        for c in all_channels:
+            c.get_context()
+
+        if periodic_exc is not None:
+            exc = periodic_exc
+            periodic_exc = None
+
+            six.reraise(exc[0], exc[1], exc[2])
+
+        run_periodic = True
+        periodic_condition.notify()
+
+
 def interact():
     """
     Called at least once per interaction.
@@ -756,60 +1030,88 @@ def interact():
     if not pcm_ok:
         return
 
-    try:
-        for c in all_channels:
+    with lock:
 
-            c.interact()
+        try:
+            for c in all_channels:
 
-            # if _music_volumes.get(i, 1.0) != c.chan_volume:
-            #    c.set_volume(_music_volumes.get(i, 1.0))
+                c.interact()
 
-            ctx = c.context
+                # if _music_volumes.get(i, 1.0) != c.chan_volume:
+                #    c.set_volume(_music_volumes.get(i, 1.0))
 
-            # If we're in the same music change, then do nothing with the
-            # music.
-            if c.last_changed == ctx.last_changed:
-                continue
+                ctx = c.context
 
-            filenames = ctx.last_filenames
-            tight = ctx.last_tight
+                # If we're in the same music change, then do nothing with the
+                # music.
+                if c.last_changed == ctx.last_changed:
+                    continue
 
-            if c.loop:
-                if not filenames or c.get_playing() not in filenames:
-                    c.fadeout(renpy.config.fade_music)
+                filenames = ctx.last_filenames
+                tight = ctx.last_tight
 
-            if filenames:
-                c.enqueue(filenames, loop=True, synchro_start=True, tight=tight)
+                if c.loop:
+                    if not filenames or c.get_playing() not in filenames:
+                        c.fadeout(renpy.config.fade_music)
 
-            c.last_changed = ctx.last_changed
+                if filenames:
+                    c.enqueue(filenames, loop=True, synchro_start=True, tight=tight)
 
-    except:
-        if renpy.config.debug_sound:
-            raise
+                c.last_changed = ctx.last_changed
+
+        except:
+            if renpy.config.debug_sound:
+                raise
 
     periodic()
 
+
 def rollback():
     """
     On rollback, we want to stop all the channels with non-empty sounds.
     """
 
-    for c in all_channels:
-        if not c.loop:
-            c.fadeout(0)
+    with lock:
+
+        for c in all_channels:
+            if not c.loop:
+                c.fadeout(0)
+
+
+global_pause = False
+
 
 def pause_all():
     """
     Pause all playback channels.
     """
 
-    for c in channels.values():
-        c.pause()
+    global global_pause
+    global_pause = True
+
+    periodic()
+
 
 def unpause_all():
     """
     Unpause all playback channels.
     """
 
-    for c in channels.values():
-        c.unpause()
+    global global_pause
+    global_pause = False
+
+    periodic()
+
+
+def sample_surfaces(rgb, rgba):
+    if not renpysound:
+        return
+
+    renpysound.sample_surfaces(rgb, rgba)
+
+
+def advance_time():
+    if not renpysound:
+        return
+
+    renpysound.advance_time()
